@@ -180,7 +180,7 @@ void MVM_VMArray_at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
             break;
         case MVM_ARRAY_I64:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos I64 expected int register");
             if (real_index >= body->elems)
                 value->i64 = 0;
             else
@@ -188,7 +188,7 @@ void MVM_VMArray_at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
             break;
         case MVM_ARRAY_I32:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos I32 expected int register");
             if (real_index >= body->elems)
                 value->i64 = 0;
             else
@@ -196,7 +196,7 @@ void MVM_VMArray_at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
             break;
         case MVM_ARRAY_I16:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos I16 expected int register");
             if (real_index >= body->elems)
                 value->i64 = 0;
             else
@@ -204,7 +204,7 @@ void MVM_VMArray_at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
             break;
         case MVM_ARRAY_I8:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos I8 expected int register");
             if (real_index >= body->elems)
                 value->i64 = 0;
             else
@@ -227,36 +227,36 @@ void MVM_VMArray_at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
                 value->n64 = (MVMnum64)body->slots.n32[body->start + real_index];
             break;
         case MVM_ARRAY_U64:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos U64 expected int register, got %d instead", kind);
             if (real_index >= body->elems)
-                value->i64 = 0;
+                value->u64 = 0;
             else
-                value->i64 = (MVMint64)body->slots.u64[body->start + real_index];
+                value->u64 = (MVMint64)body->slots.u64[body->start + real_index];
             break;
         case MVM_ARRAY_U32:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos U32 expected int register");
             if (real_index >= body->elems)
-                value->i64 = 0;
+                value->u64 = 0;
             else
-                value->i64 = (MVMint64)body->slots.u32[body->start + real_index];
+                value->u64 = (MVMint64)body->slots.u32[body->start + real_index];
             break;
         case MVM_ARRAY_U16:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos U16 expected int register");
             if (real_index >= body->elems)
-                value->i64 = 0;
+                value->u64 = 0;
             else
-                value->i64 = (MVMint64)body->slots.u16[body->start + real_index];
+                value->u64 = (MVMint64)body->slots.u16[body->start + real_index];
             break;
         case MVM_ARRAY_U8:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: atpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: atpos U8 expected int register");
             if (real_index >= body->elems)
-                value->i64 = 0;
+                value->u64 = 0;
             else
-                value->i64 = (MVMint64)body->slots.u8[body->start + real_index];
+                value->u64 = (MVMint64)body->slots.u8[body->start + real_index];
             break;
         default:
             MVM_exception_throw_adhoc(tc, "MVMArray: Unhandled slot type, got '%s'", MVM_reg_get_debug_name(tc, repr_data->slot_type));
@@ -412,22 +412,22 @@ void MVM_VMArray_bind_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
             break;
         case MVM_ARRAY_I64:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos I64 expected int register");
             body->slots.i64[body->start + real_index] = value.i64;
             break;
         case MVM_ARRAY_I32:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos I32 expected int register");
             body->slots.i32[body->start + real_index] = (MVMint32)value.i64;
             break;
         case MVM_ARRAY_I16:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos I16 expected int register");
             body->slots.i16[body->start + real_index] = (MVMint16)value.i64;
             break;
         case MVM_ARRAY_I8:
             if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos I8 expected int register");
             body->slots.i8[body->start + real_index] = (MVMint8)value.i64;
             break;
         case MVM_ARRAY_N64:
@@ -441,23 +441,23 @@ void MVM_VMArray_bind_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
             body->slots.n32[body->start + real_index] = (MVMnum32)value.n64;
             break;
         case MVM_ARRAY_U64:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos U64 expected int register");
             body->slots.u64[body->start + real_index] = value.i64;
             break;
         case MVM_ARRAY_U32:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos U32 expected int register");
             body->slots.u32[body->start + real_index] = (MVMuint32)value.i64;
             break;
         case MVM_ARRAY_U16:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos U16 expected int register");
             body->slots.u16[body->start + real_index] = (MVMuint16)value.i64;
             break;
         case MVM_ARRAY_U8:
-            if (kind != MVM_reg_int64)
-                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos expected int register");
+            if (kind != MVM_reg_uint64 && kind != MVM_reg_int64)
+                MVM_exception_throw_adhoc(tc, "MVMArray: bindpos U8 expected int register");
             body->slots.u8[body->start + real_index] = (MVMuint8)value.i64;
             break;
         default:
@@ -807,19 +807,46 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     exit_single_user(tc, body);
 }
 
+static MVMuint16 slot_type_to_kind(MVMuint8 slot_type) {
+    switch (slot_type) {
+        case MVM_ARRAY_OBJ:
+            return MVM_reg_obj;
+            break;
+        case MVM_ARRAY_STR:
+            return MVM_reg_str;
+            break;
+        case MVM_ARRAY_I64:
+        case MVM_ARRAY_I32:
+        case MVM_ARRAY_I16:
+        case MVM_ARRAY_I8:
+            return MVM_reg_int64;
+            break;
+        case MVM_ARRAY_N64:
+        case MVM_ARRAY_N32:
+            return MVM_reg_num64;
+            break;
+        case MVM_ARRAY_U64:
+        case MVM_ARRAY_U32:
+        case MVM_ARRAY_U16:
+        case MVM_ARRAY_U8:
+            return MVM_reg_uint64;
+            break;
+        default:
+            abort(); /* never reached, silence compiler warnings */
+    }
+}
+
 static void copy_elements(MVMThreadContext *tc, MVMObject *src, MVMObject *dest, MVMint64 s_offset, MVMint64 d_offset, MVMint64 elems) {
     MVMArrayBody     *s_body      = (MVMArrayBody *)OBJECT_BODY(src);
     MVMArrayBody     *d_body      = (MVMArrayBody *)OBJECT_BODY(dest);
     MVMArrayREPRData *s_repr_data = REPR(src)->ID == MVM_REPR_ID_VMArray
                                     ? (MVMArrayREPRData *)STABLE(src)->REPR_data  : NULL;
-    MVMArrayREPRData *d_repr_data = REPR(src)->ID == MVM_REPR_ID_VMArray
-                                    ? (MVMArrayREPRData *)STABLE(dest)->REPR_data : NULL;
+    MVMArrayREPRData *d_repr_data = (MVMArrayREPRData *)STABLE(dest)->REPR_data;
 
     if (elems > 0) {
         MVMint64  i;
-        MVMuint16 kind;
         MVMuint8 d_needs_barrier = dest->header.flags2 & MVM_CF_SECOND_GEN;
-        if (s_repr_data && d_repr_data
+        if (s_repr_data
                 && s_repr_data->slot_type == d_repr_data->slot_type
                 && s_repr_data->elem_size == d_repr_data->elem_size
                 && (d_repr_data->slot_type != MVM_ARRAY_OBJ || !d_needs_barrier)
@@ -833,36 +860,13 @@ static void copy_elements(MVMThreadContext *tc, MVMObject *src, MVMObject *dest,
             );
         }
         else {
-            switch (s_repr_data->slot_type) {
-                case MVM_ARRAY_OBJ:
-                    kind = MVM_reg_obj;
-                    break;
-                case MVM_ARRAY_STR:
-                    kind = MVM_reg_str;
-                    break;
-                case MVM_ARRAY_I64:
-                case MVM_ARRAY_I32:
-                case MVM_ARRAY_I16:
-                case MVM_ARRAY_I8:
-                    kind = MVM_reg_int64;
-                    break;
-                case MVM_ARRAY_N64:
-                case MVM_ARRAY_N32:
-                    kind = MVM_reg_num64;
-                    break;
-                case MVM_ARRAY_U64:
-                case MVM_ARRAY_U32:
-                case MVM_ARRAY_U16:
-                case MVM_ARRAY_U8:
-                    kind = MVM_reg_int64;
-                    break;
-                default:
-                    abort(); /* never reached, silence compiler warnings */
-            }
+            MVMuint16 target_kind = slot_type_to_kind(d_repr_data->slot_type);
+            MVMuint16 source_kind = slot_type_to_kind(s_repr_data->slot_type);
             for (i = 0; i < elems; i++) {
                 MVMRegister to_copy;
-                REPR(src)->pos_funcs.at_pos(tc, STABLE(src), src, s_body, s_offset + i, &to_copy, kind);
-                MVM_VMArray_bind_pos(tc, STABLE(dest), dest, d_body, d_offset + i, to_copy, kind);
+                REPR(src)->pos_funcs.at_pos(tc, STABLE(src), src, s_body, s_offset + i, &to_copy, source_kind);
+                /* actually should coerce between source_kind and target_kind here */
+                MVM_VMArray_bind_pos(tc, STABLE(dest), dest, d_body, d_offset + i, to_copy, target_kind);
             }
         }
     }
@@ -871,8 +875,7 @@ static void copy_elements(MVMThreadContext *tc, MVMObject *src, MVMObject *dest,
 static void aslice(MVMThreadContext *tc, MVMSTable *st, MVMObject *src, void *data, MVMObject *dest, MVMint64 start, MVMint64 end) {
     MVMArrayBody     *s_body      = (MVMArrayBody *)data;
     MVMArrayBody     *d_body      = (MVMArrayBody *)OBJECT_BODY(dest);
-    MVMArrayREPRData *d_repr_data = REPR(dest)->ID == MVM_REPR_ID_VMArray
-                                      ? STABLE(dest)->REPR_data : NULL;
+    MVMArrayREPRData *d_repr_data = STABLE(dest)->REPR_data;
 
     MVMint64 total_elems = REPR(src)->elems(tc, st, src, s_body);
     MVMint64 elems;
@@ -884,10 +887,7 @@ static void aslice(MVMThreadContext *tc, MVMSTable *st, MVMObject *src, void *da
     }
 
     elems = end - start + 1;
-    if (d_repr_data) {
-        set_size_internal(tc, d_body, elems, d_repr_data);
-    }
-
+    set_size_internal(tc, d_body, elems, d_repr_data);
     copy_elements(tc, src, dest, start, 0, elems);
 }
 
@@ -897,6 +897,10 @@ static void write_buf(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
     MVMuint64 start = body->start;
     MVMuint64 elems = body->elems;
 
+    /* Throw on invalid slot type */
+    if (repr_data->slot_type < MVM_ARRAY_I64) {
+        MVM_exception_throw_adhoc(tc, "MVMArray: write_buf requires an integer type");
+    }
     /* Throw on negative offset. */
     if (offset < 0) {
         MVM_exception_throw_adhoc(tc, "MVMArray: Index out of bounds");
@@ -911,13 +915,18 @@ static void write_buf(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
     memcpy(body->slots.u8 + (start + offset) * repr_data->elem_size, from, count);
 }
 
-MVMint64 read_buf(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 offset, MVMuint64 count) {
+static MVMint64 read_buf(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 offset, MVMuint64 count) {
     MVMArrayREPRData *repr_data = (MVMArrayREPRData *)st->REPR_data;
     MVMArrayBody     *body      = (MVMArrayBody *)data;
     MVMint64 start = body->start;
     MVMint64 result = 0;
     size_t elem_size = repr_data->elem_size;
 
+    /* Throw on invalid slot type */
+    if (repr_data->slot_type < MVM_ARRAY_I64) {
+        MVM_exception_throw_adhoc(tc, "MVMArray: read_buf requires an integer type");
+    }
+
     if (offset < 0 || (start + body->elems) * elem_size < offset * elem_size + count) {
         MVM_exception_throw_adhoc(tc, "MVMArray: read_buf out of bounds offset %"PRIi64" start %"PRIi64" elems %"PRIu64" count %"PRIu64, offset, start, body->elems, count);
     }
@@ -1073,7 +1082,7 @@ static MVMStorageSpec get_elem_storage_spec(MVMThreadContext *tc, MVMSTable *st)
         case MVM_ARRAY_U16:
         case MVM_ARRAY_U8:
             spec.inlineable      = MVM_STORAGE_SPEC_INLINED;
-            spec.boxed_primitive = MVM_STORAGE_SPEC_BP_INT;
+            spec.boxed_primitive = MVM_STORAGE_SPEC_BP_UINT64;
             spec.can_box         = MVM_STORAGE_SPEC_CAN_BOX_INT;
             spec.is_unsigned     = 1;
             break;
@@ -1118,6 +1127,7 @@ static AO_t * pos_as_atomic_multidim(MVMThreadContext *tc, MVMSTable *st, MVMObj
 /* Compose the representation. */
 static void spec_to_repr_data(MVMThreadContext *tc, MVMArrayREPRData *repr_data, const MVMStorageSpec *spec) {
     switch (spec->boxed_primitive) {
+        case MVM_STORAGE_SPEC_BP_UINT64:
         case MVM_STORAGE_SPEC_BP_INT:
             if (spec->is_unsigned) {
                 switch (spec->bits) {
@@ -1265,47 +1275,57 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
     if (body->ssize)
         body->slots.any = MVM_malloc(body->ssize * repr_data->elem_size);
 
-    for (i = 0; i < body->elems; i++) {
-        switch (repr_data->slot_type) {
-            case MVM_ARRAY_OBJ:
+    switch (repr_data->slot_type) {
+        case MVM_ARRAY_OBJ:
+            for (i = 0; i < body->elems; i++)
                 MVM_ASSIGN_REF(tc, &(root->header), body->slots.o[i], MVM_serialization_read_ref(tc, reader));
-                break;
-            case MVM_ARRAY_STR:
+            break;
+        case MVM_ARRAY_STR:
+            for (i = 0; i < body->elems; i++)
                 MVM_ASSIGN_REF(tc, &(root->header), body->slots.s[i], MVM_serialization_read_str(tc, reader));
-                break;
-            case MVM_ARRAY_I64:
+            break;
+        case MVM_ARRAY_I64:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i64[i] = MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_I32:
+            break;
+        case MVM_ARRAY_I32:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i32[i] = (MVMint32)MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_I16:
+            break;
+        case MVM_ARRAY_I16:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i16[i] = (MVMint16)MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_I8:
+            break;
+        case MVM_ARRAY_I8:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i8[i] = (MVMint8)MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_U64:
+            break;
+        case MVM_ARRAY_U64:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i64[i] = MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_U32:
+            break;
+        case MVM_ARRAY_U32:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i32[i] = (MVMuint32)MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_U16:
+            break;
+        case MVM_ARRAY_U16:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i16[i] = (MVMuint16)MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_U8:
+            break;
+        case MVM_ARRAY_U8:
+            for (i = 0; i < body->elems; i++)
                 body->slots.i8[i] = (MVMuint8)MVM_serialization_read_int(tc, reader);
-                break;
-            case MVM_ARRAY_N64:
+            break;
+        case MVM_ARRAY_N64:
+            for (i = 0; i < body->elems; i++)
                 body->slots.n64[i] = MVM_serialization_read_num(tc, reader);
-                break;
-            case MVM_ARRAY_N32:
+            break;
+        case MVM_ARRAY_N32:
+            for (i = 0; i < body->elems; i++)
                 body->slots.n32[i] = (MVMnum32)MVM_serialization_read_num(tc, reader);
-                break;
-            default:
-                MVM_exception_throw_adhoc(tc, "MVMArray: Unhandled slot type");
-        }
+            break;
+        default:
+            MVM_exception_throw_adhoc(tc, "MVMArray: Unhandled slot type");
     }
 }
 
@@ -1498,6 +1518,14 @@ void *MVM_VMArray_find_fast_impl_for_jit(MVMThreadContext *tc, MVMSTable *st, MV
                 return vmarray_at_pos_int64;
             }
             break;
+        case MVM_OP_atpos_u:
+            if (kind != MVM_reg_uint64) {
+                return NULL;
+            }
+            if (repr_data->slot_type == MVM_ARRAY_U64) {
+                return vmarray_at_pos_int64;
+            }
+            break;
         case MVM_OP_bindpos_i:
             if (kind != MVM_reg_int64) {
                 return NULL;
@@ -1506,6 +1534,14 @@ void *MVM_VMArray_find_fast_impl_for_jit(MVMThreadContext *tc, MVMSTable *st, MV
                 return vmarray_bind_pos_int64;
             }
             break;
+        case MVM_OP_bindpos_u:
+            if (kind != MVM_reg_uint64) {
+                return NULL;
+            }
+            if (repr_data->slot_type == MVM_ARRAY_U64) {
+                return vmarray_bind_pos_int64;
+            }
+            break;
         default:
             return NULL;
     }
